(in-package "MORPHICL")

(defclass input-morph (morph)
  ((input             :type (array character)
                      :initform (make-array 32 
                                            :element-type 'character 
                                            :adjustable t 
                                            :fill-pointer 0)
                      :accessor input-morph-input)
   (on-activate       :type (or function null)
                      :initarg :on-activate
                      :initform nil
                      :accessor input-morph-on-activate)
   (history           :type (or (array character) null)
                      :initarg :history
                      :initform (make-array 32
                                            :element-type 'string
                                            :adjustable t
                                            :fill-pointer 0)
                      :accessor input-morph-history)
   (history-pointer   :type integer
                      :initform 0
                      :accessor input-morph-history-pointer)
   (insertion-pointer :type integer
                      :initform 0
                      :accessor input-morph-insertion-pointer)
   (display-pointer   :type integer
                      :initform 0
                      :accessor input-morph-display-pointer)))

(defmethod morph-draw ((morph input-morph) (canvas canvas))
  (when (>= (- (input-morph-insertion-pointer morph)
               (input-morph-display-pointer morph))
            (round (rectangle-width (canvas-transformation canvas))
                   (font-width (canvas-font canvas))))
    (setf (input-morph-display-pointer morph)
          (- (1+ (input-morph-insertion-pointer morph))
             (round (rectangle-width (canvas-transformation canvas))
                    (font-width (canvas-font canvas))))))
  (with-canvas-options (canvas :foreground-color (morph-highlight-color morph))
    (canvas-fill-rectangle canvas
                           (canvas-translate-rectangle canvas
                             (make-rectangle (* (font-width (canvas-font canvas))
                                                (- (input-morph-insertion-pointer morph)
                                                   (input-morph-display-pointer morph)))
                                             0
                                             (* (font-width (canvas-font canvas))
                                                (- (1+ (input-morph-insertion-pointer morph))
                                                   (input-morph-display-pointer morph)))
                                             (font-height (canvas-font canvas))))))
    (canvas-draw-text canvas
                      (input-morph-input morph)
                      (canvas-transform-point canvas (make-point 0 0))
                      :start (input-morph-display-pointer morph)
                      :end (fill-pointer (input-morph-input morph))))

(defmethod input-morph-backspace ((morph input-morph))
  (unless (zerop (input-morph-insertion-pointer morph))
    (loop for position from (1- (input-morph-insertion-pointer morph))
                       below (1- (fill-pointer (input-morph-input morph)))
      do (setf (aref (input-morph-input morph) position)
               (aref (input-morph-input morph) (1+ position))))
    (when (<= (input-morph-insertion-pointer morph)
              (input-morph-display-pointer morph))
      (decf (input-morph-display-pointer morph)))
    (decf (input-morph-insertion-pointer morph))
    (decf (fill-pointer (input-morph-input morph)))))

(defmethod input-morph-rubout ((morph input-morph))
  (when (< (input-morph-insertion-pointer morph)
           (fill-pointer (input-morph-input morph)))
    (loop for position from (input-morph-insertion-pointer morph)
                       below (1- (fill-pointer (input-morph-input morph)))
      do (setf (aref (input-morph-input morph) position)
               (aref (input-morph-input morph) (1+ position))))
    (decf (fill-pointer (input-morph-input morph)))))

(defmethod input-morph-cursor-home ((morph input-morph))
  (setf (input-morph-display-pointer morph) 0)
  (setf (input-morph-insertion-pointer morph) 0))

(defmethod input-morph-cursor-end ((morph input-morph))
  (setf (input-morph-display-pointer morph) 0)
  (setf (input-morph-insertion-pointer morph)
        (fill-pointer (input-morph-input morph))))

(defmethod input-morph-cursor-left ((morph input-morph))
  (when (> (input-morph-insertion-pointer morph) 0)
    (when (<= (input-morph-insertion-pointer morph)
              (input-morph-display-pointer morph))
      (decf (input-morph-display-pointer morph)))
    (decf (input-morph-insertion-pointer morph))))

(defmethod input-morph-cursor-right ((morph input-morph))
  (when (< (input-morph-insertion-pointer morph)
           (fill-pointer (input-morph-input morph)))
    (incf (input-morph-insertion-pointer morph))))

(defmethod input-morph-flush-input ((morph input-morph))
  (when (and (input-morph-history morph)
             (> (fill-pointer (input-morph-input morph)) 0)
             (or (zerop (fill-pointer (input-morph-history morph)))
                 (not (equal (input-morph-input morph)
                             (aref (input-morph-history morph)
                                   (input-morph-history-pointer morph))))))
    (vector-push-extend (make-array (fill-pointer (input-morph-input morph))
                                    :element-type 'character
                                    :initial-contents (input-morph-input morph))
                        (input-morph-history morph)))
  (setf (input-morph-display-pointer morph) 0)
  (setf (input-morph-insertion-pointer morph) 0)
  (setf (fill-pointer (input-morph-input morph)) 0))

(defmethod input-morph-restore-history ((morph input-morph))
  (let ((history-item (aref (input-morph-history morph)
                            (input-morph-history-pointer morph))))
    (dotimes (position (length history-item))
      (setf (aref (input-morph-input morph) position)
            (char history-item position)))
    (setf (input-morph-display-pointer morph) 0)
    (setf (input-morph-insertion-pointer morph) (length history-item))
    (setf (fill-pointer (input-morph-input morph)) (length history-item))))

(defmethod input-morph-cursor-down ((morph input-morph))
  (when (and (input-morph-history morph)
             (> (fill-pointer (input-morph-history morph)) 0))
    (input-morph-flush-input morph)
    (incf (input-morph-history-pointer morph))
    (when (= (input-morph-history-pointer morph)
             (fill-pointer (input-morph-history morph)))
      (setf (input-morph-history-pointer morph) 0))
    (input-morph-restore-history morph)))

(defmethod input-morph-cursor-up ((morph input-morph))
  (when (and (input-morph-history morph)
             (> (fill-pointer (input-morph-history morph)) 0))
    (input-morph-flush-input morph)
    (when (zerop (input-morph-history-pointer morph))
      (setf (input-morph-history-pointer morph)
            (fill-pointer (input-morph-history morph))))
    (decf (input-morph-history-pointer morph))
    (input-morph-restore-history morph)))

(defmethod input-morph-return ((morph input-morph))
  (when (input-morph-on-activate morph)
    (funcall (input-morph-on-activate morph)
             morph
             (make-array (fill-pointer (input-morph-input morph))
                         :element-type 'character
                         :initial-contents (input-morph-input morph))))
  (input-morph-flush-input morph))

(defmethod input-morph-insert-character ((morph input-morph) character)
  (cond
    ((< (input-morph-insertion-pointer morph)
        (fill-pointer (input-morph-input morph)))
       (vector-push-extend #\Space (input-morph-input morph))
       (loop for position from (1- (fill-pointer (input-morph-input morph)))
                          downto (1+ (input-morph-insertion-pointer morph))
         do (setf (aref (input-morph-input morph) position)
                  (aref (input-morph-input morph) (1- position))))
       (setf (aref (input-morph-input morph)
                   (input-morph-insertion-pointer morph))
             character))
    (t
       (vector-push-extend character
                           (input-morph-input morph))))
  (incf (input-morph-insertion-pointer morph)))

(defmethod morph-handle-event ((morph input-morph) (event key-press-event))
  (case (key-event-key event)
    (#\Backspace
      (input-morph-backspace morph))
    (#\Rubout
      (input-morph-rubout morph))
    (#\Return
      (input-morph-return morph))
    (:cursor-left
      (input-morph-cursor-left morph))
    (:cursor-right
      (input-morph-cursor-right morph))
    (:cursor-up
      (input-morph-cursor-up morph))
    (:cursor-down
      (input-morph-cursor-down morph))
    (:cursor-home
      (input-morph-cursor-home morph))
    (:cursor-end
      (input-morph-cursor-end morph))
    (otherwise
      (when (characterp (key-event-key event))
        (input-morph-insert-character morph
                                      (key-event-key event)))))
  (morph-full-draw (port-world (event-port event))
                   (event-port event))
  event)

