(in-package "MORPHICL")

(defclass rectangle ()
  ((left         :type number  :initarg :left          :accessor rectangle-left)
   (top          :type number  :initarg :top           :accessor rectangle-top)
   (right        :type number  :initarg :right         :accessor rectangle-right)
   (bottom       :type number  :initarg :bottom        :accessor rectangle-bottom)))

(defun rectanglep (object)
  (typep object 'rectangle))

(defgeneric rectangle-top ((rectangle rectangle)))
(defgeneric (setf rectangle-top) (top (rectangle rectangle)))
(defgeneric rectangle-bottom ((rectangle rectangle)))
(defgeneric (setf rectangle-bottom) (bottom (rectangle rectangle)))
(defgeneric rectangle-left ((rectangle rectangle)))
(defgeneric (setf rectangle-left) (left (rectangle rectangle)))
(defgeneric rectangle-right ((rectangle rectangle)))
(defgeneric (setf rectangle-right) (right (rectangle rectangle)))
(defgeneric rectangle-top-left ((rectangle rectangle)))
(defgeneric* (setf rectangle-top-left) (left top (rectangle rectangle)))
(defgeneric rectangle-top-right ((rectangle rectangle)))
(defgeneric* (setf rectangle-top-right) (right top (rectangle rectangle)))
(defgeneric rectangle-bottom-left ((rectangle rectangle)))
(defgeneric* (setf rectangle-bottom-left) (left bottom (rectangle rectangle)))
(defgeneric rectangle-bottom-right ((rectangle rectangle)))
(defgeneric* (setf rectangle-bottom-right) (right bottom (rectangle rectangle)))
(defgeneric rectangle-bounds ((rectangle rectangle)))
(defgeneric* (setf rectangle-bounds) (left top right bottom (rectangle rectangle)))

(defgeneric rectangle-round ((rectangle rectangle)))
(defgeneric rectangle-translate ((rectangle rectangle) x y))
(defgeneric rectangle-scale ((rectangle rectangle) x y))
(defgeneric rectangle-intersect ((a rectangle) (b rectangle)))
(defgeneric rectangle-union ((a rectangle) (b rectangle)))
(defgeneric rectangle-height ((rectangle rectangle)))
(defgeneric rectangle-width ((rectangle rectangle)))
(defgeneric rectangle-size ((rectangle rectangle)))
(defgeneric rectangle-area ((rectangle rectangle)))
(defgeneric rectangle-center ((rectangle rectangle)))

(defun make-rectangle (left top right bottom)
  (make-instance 'rectangle
                 :left left
                 :top top
                 :right right
                 :bottom bottom))

(defun copy-rectangle (rectangle)
  (make-instance 'rectangle
                 :left (rectangle-left rectangle)
                 :top (rectangle-top rectangle)
                 :right (rectangle-right rectangle)
                 :bottom (rectangle-bottom rectangle)))

(defmethod rectangle-top-left ((rectangle rectangle))
  (values (rectangle-left rectangle) (rectangle-top rectangle)))

(defmethod* (setf rectangle-top-left) (top left (rectangle rectangle))
  (values (setf (rectangle-left rectangle) left)
          (setf (rectangle-top rectangle) top)))

(defmethod rectangle-top-right ((rectangle rectangle))
  (values (rectangle-right rectangle) (rectangle-top rectangle)))

(defmethod* (setf rectangle-top-right) (top right (rectangle rectangle))
  (values (setf (rectangle-right rectangle) right)
          (setf (rectangle-top rectangle) top)))

(defmethod rectangle-bottom-left ((rectangle rectangle))
  (values (rectangle-left rectangle) (rectangle-bottom rectangle)))

(defmethod* (setf rectangle-bottom-left) (bottom left (rectangle rectangle))
  (values (setf (rectangle-left rectangle) left)
          (setf (rectangle-bottom rectangle) bottom)))

(defmethod rectangle-bottom-right ((rectangle rectangle))
  (values (rectangle-right rectangle) (rectangle-bottom rectangle)))

(defmethod* (setf rectangle-bottom-right) (bottom right (rectangle rectangle))
  (values (setf (rectangle-right rectangle) right)
          (setf (rectangle-bottom rectangle) bottom)))

(defmethod rectangle-bounds ((rectangle rectangle))
  (values (rectangle-left rectangle)
          (rectangle-top rectangle)
          (rectangle-right rectangle)
          (rectangle-bottom rectangle)))

(defmethod* (setf rectangle-bounds) (left top right bottom (rectangle rectangle))
  (values (setf (rectangle-left rectangle) left)
          (setf (rectangle-top rectangle) top)
          (setf (rectangle-right rectangle) right)
          (setf (rectangle-bottom bottom) bottom)))

(defmethod rectangle-round ((rectangle rectangle))
  (setf (rectangle-left rectangle) (round (rectangle-left rectangle)))
  (setf (rectangle-top rectangle) (round (rectangle-top rectangle)))
  (setf (rectangle-right rectangle) (round (rectangle-right rectangle)))
  (setf (rectangle-bottom rectangle) (round (rectangle-bottom rectangle)))
  rectangle)

(defmethod rectangle-translate ((rectangle rectangle) x y)
  (incf (rectangle-left rectangle) x)
  (incf (rectangle-top rectangle) y)
  (incf (rectangle-right rectangle) x)
  (incf (rectangle-bottom rectangle) y)
  rectangle)

(defmethod rectangle-scale ((rectangle rectangle) x y)
  (setf (rectangle-left rectangle) (* x (rectangle-left rectangle)))
  (setf (rectangle-top rectangle) (* y (rectangle-top rectangle)))
  (setf (rectangle-right rectangle) (* x (rectangle-right rectangle)))
  (setf (rectangle-bottom rectangle) (* y (rectangle-bottom rectangle)))
  rectangle)

(defmethod rectangle-intersect ((a rectangle) (b rectangle))
  (setf (rectangle-left a)
        (max (rectangle-left a) (rectangle-left b)))
  (setf (rectangle-top a)
        (max (rectangle-top a) (rectangle-top b)))
  (setf (rectangle-right a)
        (min (rectangle-right a) (rectangle-right b)))
  (setf (rectangle-bottom a)
        (min (rectangle-bottom a) (rectangle-bottom b)))
  a)

(defmethod rectangle-union ((a rectangle) (b rectangle))
  (setf (rectangle-left a)
        (min (rectangle-left a) (rectangle-left b)))
  (setf (rectangle-top a)
        (min (rectangle-top a) (rectangle-top b)))
  (setf (rectangle-right a)
        (max (rectangle-right a) (rectangle-right b)))
  (setf (rectangle-bottom a)
        (max (rectangle-bottom a) (rectangle-bottom b)))
  a)

(defmethod rectangle-width ((rectangle rectangle))
  (- (rectangle-right rectangle) (rectangle-left rectangle)))

(defmethod rectangle-height ((rectangle rectangle))
  (- (rectangle-bottom rectangle) (rectangle-top rectangle)))

(defmethod rectangle-size ((rectangle rectangle))
  (values (rectangle-width rectangle)
          (rectangle-height rectangle)))

(defmethod rectangle-area ((rectangle rectangle))
  (* (rectangle-width rectangle) (rectangle-height rectangle)))

(defmethod rectangle-center ((rectangle rectangle))
  (values (/ (+ (rectangle-left rectangle) (rectangle-right rectangle)) 2)
          (/ (+ (rectangle-top rectangle) (rectangle-bottom rectangle)) 2)))

